/*
 * fft2pc.c -- Character set translator for Final Fantasy Tactics by
 *	Squaresoft.  This is not intended for copyright infringement.
 *
 * Copyright 2000, George Greer <greerga@muc.muohio.edu>
 *
 * Licensed under the GNU General Public License (GPL)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include "mmaplib.h"

static const char *program = __FILE__ ": Final Fantasy Tactics character set translator.";
static const char *version = "v0.4 (Build: "__DATE__" "__TIME__")";
static const char *copyright = "Copyright 2000, George Greer <greerga@muc.muohio.edu>";
static const char *license = "Licensed under GNU General Public License (GPL) v2 or later.";

#define FALSE	(0)
#define TRUE	(!FALSE)

struct options_s {
  unsigned int rawmultibyte : 1;
};

/* ------------------------------------------------------------------------------------------------------------------------------ */

struct {
  unsigned char match;
  struct {
    unsigned char multibyte;
    const char *text;
  } submatch[256];			/* Yeah, well, :P */
  const char *text;
} xlate[] = {
  { 0x3E, {},	"!"		},
  { 0x40, {},	"?"		},
  { 0x42, {},	"+"		},
  { 0x44, {},	"/"		},
  { 0x46, {},	":"		},
  { 0x5F, {},	"."		},
  { 0x8B, {},	"."		},	/* Really a centered . */
  { 0x8D, {},	"("		},
  { 0x8E, {},	")"		},
  /* 0x91-0xA0 Hiragana/Katakana from Gelwan on www.ffnet.net boards. */
  { 0x91, {},	"\""		},	/* Yet another one. */
  { 0x92, {},	"a"		},
  { 0x93, {},	"'"		},	/* One ' isn't enough? (See job results.) */
  { 0x94, {},	"i"		},
  { 0x95, {},	"xu"		},
  { 0x96, {},	"u"		},
  { 0x97, {},	"xe"		},
  { 0x98, {},	"e"		},
  { 0x99, {},	"xo"		},
  { 0x9A, {},	"o"		},
  { 0x9B, {},	"ka"		},
  { 0x9C, {},	"ga"		},
  { 0x9D, {},	"ki"		},
  { 0x9E, {},	"gi"		},
  { 0x9F, {},	"ku"		},
  { 0xA0, {},	"gu"		},
  { 0xD1, {
    { 0x10,	"@"		},
/*  { 0x17,	"???"		},	"<E1>'s <EB> <D1/17><E4>!!" */
    { 0x18,	"["		},
    { 0x1A,	"!"		},
/*  { 0x1B,	???		},	"We left Goug Machine City[D1/1B]" -- perhaps '...' ? */
    { 0x1D,	"-"		},
    { 0x1E,	"+"		},
    { 0x1F,	"[X]"		},	/* The 'X' controller button. */
    { 0x00,	NULL		},	/* GUARD */
  }, NULL },
  { 0xD9, {
/*  { 0xB5,	"???"		},	"<FF><D9/B5><FF>" (at end of menu) */
    { 0xB6,	"."		},
    { 0xB8,	"%"		},
    { 0xB9,	"[O]"		},	/* The 'O' controller button. */
    { 0xBA,	"<="		},	/* Left arrow on name input screen. */
    { 0xBB,	"=>"		},	/* Right arrow on name input screen. */
    { 0xBC,	"."		},	/* Really a . in center of character space. */
    { 0xBD,	":"		},
    { 0xBE,	"("		},
    { 0xBF,	")"		},
    { 0xC0,	"\""		},
    { 0xC1,	"'"		},
    { 0xC4,	"]"		},
    { 0xC6,	"/"		},
    { 0xC7,	"[T]"		},	/* The 'Triangle' controller button. */
    { 0xC8,	"[S]"		},	/* The 'Square' controller button. */
    { 0xC9,	"?"		},
    { 0x00,	NULL		},	/* GUARD */
  }, NULL },
  { 0xDA, {
    /* FIXME: Do these Zodiac Stone symbol listings correlate properly? */
    { 0x00,	"{Cancer}"	},	/* Zodiac Stone symbol. */
    { 0x01,	"{Leo}"		},	/* Zodiac Stone symbol. */
    { 0x02,	"{Virgo}"	},	/* Zodiac Stone symbol. */
    { 0x03,	"{Libra}"	},	/* Zodiac Stone symbol. */
    { 0x04,	"{Scorpio}"	},	/* Zodiac Stone symbol. */
    { 0x05,	"{Sagittarius}"	},	/* Zodiac Stone symbol. */
    { 0x06,	"{Capricorn}"	},	/* Zodiac Stone symbol. */
    { 0x07,	"{Aquarius}"	},	/* Zodiac Stone symbol. */
    { 0x08,	"{Pisces}"	},	/* Zodiac Stone symbol. */
    { 0x09,	"{Aries}"	},	/* Zodiac Stone symbol. */
    { 0x0A,	"{Taurus}"	},	/* Zodiac Stone symbol. */
    { 0x0B,	"{Gemini}"	},	/* Zodiac Stone symbol. */
/*  { 0x70	"???"		},	"Damage <DA/70> target's MP" */
    { 0x73,	" "		},	/* Really?? */
    { 0x74,	","		},
    { 0x75,	":"		},
    { 0x76,	"'"		},	/* Three ' now. Sheesh. */
    { 0x77,	"\""		},	/* Another one? */
    { 0x00,	NULL		},	/* GUARD */
  }, NULL },
  { 0xE3, {
    { 0x00,	"[Normal Text]"	},
    { 0x04,	"[Gray Text]"	},
    { 0x08,	"[Red Text]"	},
    { 0x00,	NULL		},	/* GUARD */
  }, NULL },
  { 0xE0, {},	"[Ramza]"	},
  { 0xE1, {},	"[Variable: 1]"	},
  { 0xE2, {},	"[Variable: 2]"	},
  { 0xE3, {},	"[Variable: 3]"	},
  { 0xE4, {},	"[Variable: 4]"	},
  { 0xE5, {},	"[Variable: 5]"	},
  { 0xE6, {},	"[Variable: 6]"	},
  { 0xE7, {
    { 0x02,	"  "		},	/* Name input screen uses for spacing. */
    { 0x03,	"  "		},	/* Name input screen uses for spacing. */
    { 0x04,	"  "		},	/* Name input screen uses for spacing. */
    { 0x16,	"  "		},	/* Name input screen uses for spacing. */
    { 0x17,	"  "		},	/* Name input screen uses for spacing. */
    { 0x2A,	"  "		},	/* Name input screen uses for spacing. */
    { 0x2B,	"  "		},	/* Name input screen uses for spacing. */
    { 0x3E,	"  "		},	/* Name input screen uses for spacing. */
    { 0x3F,	"  "		},	/* Name input screen uses for spacing. */
    { 0x52,	"  "		},	/* Name input screen uses for spacing. */
    { 0x53,	"  "		},	/* Name input screen uses for spacing. */
    { 0x66,	"  "		},	/* Name input screen uses for spacing. */
    { 0x67,	"  "		},	/* Name input screen uses for spacing. */
    { 0x68,	"  "		},	/* Name input screen uses for spacing. */
    { 0x78,	"  "		},	/* Name input screen uses for spacing. */
    { 0x79,	"  "		},	/* Name input screen uses for spacing. */
    { 0x7A,	"  "		},	/* Name input screen uses for spacing. */
    { 0x7B,	"  "		},	/* Name input screen uses for spacing. */
    { 0x8E,	"  "		},	/* Name input screen uses for spacing. */
    { 0x8F,	"  "		},	/* Name input screen uses for spacing. */
    { 0xA2,	"  "		},	/* Name input screen uses for spacing. */
    { 0xA3,	"  "		},	/* Name input screen uses for spacing. */
    { 0xA4,	"  "		},	/* Name input screen uses for spacing. */
    { 0xB6,	"  "		},	/* Name input screen uses for spacing. */
    { 0xB7,	"  "		},	/* Name input screen uses for spacing. */
    { 0xC9,	"  "		},	/* Name input screen uses for spacing. */
    { 0xCA,	"  "		},	/* Name input screen uses for spacing. */
    { 0x00,	NULL		},	/* GUARD */
  }, NULL	},
  { 0xE8, {},	"[Variable: 8]"	},
  { 0xE9, {},	"[Variable: 9]"	},
  { 0xEA, {},	"[Variable: A]"	},
  { 0xEB, {},	"[Variable: B]"	},
  { 0xEC, {},	"[<special>]"	},	/* Battle character Brave Story profile picture number. */
  { 0xF0, {},	NULL		},	/* Compression keys. */
  { 0xF5, {},	NULL		},	/* Used for Brave Story pictures. */
  { 0xF6, {},	NULL		},	/* Used for Brave Story pictures. */
  { 0xF8, {},	"\n"		},	/* In-text */
  { 0xFA, {},	" "		},
/*{ 0xFD, {},	"???"		},	"Carve <FD> Model" */
  { 0xFE, {},	"\n\n"		},	/* End-of-text */
  { 0x00, {},	NULL		},	/* GUARD */
};

int fft2pc(const unsigned char *input, int length, struct options_s *options)
{
  int rpos;

  for (rpos = 0; rpos < length; rpos++) {
    if (input[rpos] <= 0x9)
      putchar('0' + input[rpos]);
    else if (input[rpos] <= 0x23)
      putchar('A' + (input[rpos] - 0x0A));
    else if (input[rpos] <= 0x3D)
      putchar('a' + (input[rpos] - 0x24));
    else if (input[rpos] == 0xEC)
      printf("[Picture #%u]", input[++rpos]);
    else {
      int first, second, found = FALSE;

      for (first = 0; xlate[first].text || xlate[first].match; first++) {
        if (input[rpos] != xlate[first].match)
          continue;

        if (options->rawmultibyte) {
          if (xlate[first].text)
            printf("<%0X>", input[rpos]);
          else {
            rpos++;
            printf("<%0X/%0X>", input[rpos - 1], input[rpos]);
          }
        } else if (xlate[first].text)
          fputs(xlate[first].text, stdout);
        else if (xlate[first].submatch) {
          rpos++;
          for (second = 0; xlate[first].submatch[second].text; second++) {
            if (input[rpos] != xlate[first].submatch[second].multibyte)
              continue;
            fputs(xlate[first].submatch[second].text, stdout);
            found = TRUE;
            break;
          }
          if (!found && (found = TRUE))
            printf("<%0X/%0X>", input[rpos - 1], input[rpos]);
        }
        found = TRUE;
        break;
      }
      if (!found && (found = TRUE))
        printf("<%0X>", input[rpos]);
    } /* if */
  } /* for */
  return rpos;
}

/* ------------------------------------------------------------------------------------------------------------------------------ */

struct option /* {
  const char *name;
  int has_arg;
  int *flag;
  int val;
} */ long_options[] = {
  { "help",     FALSE, NULL, 0 },
  { "version",  FALSE, NULL, 1 },
  { 0, 0, 0, 0 }
};

int print_help(char *progname)
{
  printf("%s [-r] filename\n"
	 "Use -V or --version for version and copyright information.\n"
	 "%s --help	This screen.\n",
	progname, progname);
  return EXIT_FAILURE;
}

int print_version(char *progname)
{
  printf("%s\n\t%s\n\t%s\n\n%s: version %s\n", program, copyright, license, progname, version);
  return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
  void *handle;
  char *filename;
  int gto;
  struct options_s options = {
    rawmultibyte:	FALSE,
  };

  while ((gto = getopt_long(argc, argv, "hrV", long_options, NULL)) >= 0) {
    switch (gto) {
    case 0:
    case 'h':
      return print_help(argv[0]);
    case 1:
    case 'V':
      return print_version(argv[0]);
    case 'r':
      options.rawmultibyte = TRUE;
      break;

    case ':':
      fprintf(stderr, "Option -%c requires an extra argument.\n", optopt);
      return EXIT_FAILURE;
    case '?':
      fprintf(stderr, "Unknown option -%c.\n", optopt);
      return EXIT_FAILURE;
    default:
      fputs("Unknown error condition.\n", stderr);
      return EXIT_FAILURE;
    }
  }

  if (argc == optind)
    fputs("Program requires a filename to translate.\nTry '--help' for more information.\n", stderr);
  else {
    filename = argv[optind];
    if (!(handle = mapopen(filename)))
      error(filename);
    fft2pc(mapdata(handle), maplength(handle), &options);
    mapclose(handle);
    return EXIT_SUCCESS;
  }
  return EXIT_FAILURE;
}
